#!/usr/bin/perl -wT
# 
# tinygraph -- a RRDtool frontend for tinydns statistics
# Copyright (C) 2005    Ulrich Zehl <ulrich@topfen.net>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use RRDs;
use Time::TAI64 qw(tai2unix);
use Getopt::Long;
use File::Tail;

# Configuration
my $rrd            = 'tinygraph.rrd';
my $logfile        = '/var/log/tinydns/current';
my $rrdstep = 60;
my $xpoints = 540;
my $points_per_sample = 3;

# DNS query types as logged by tinydns
my %querytypes = (
	'0001' => 'A',
	'0002' => 'NS',
	'0005' => 'CNAME',
	'0006' => 'SOA',
	'000c' => 'PTR',
	'000d' => 'HINFO',
	'000f' => 'MX',
	'0010' => 'TXT',
	'0011' => 'RP',
	'0018' => 'SIG',
	'0019' => 'KEY',
	'001c' => 'AAAA',
	'0026' => 'A6',
	'00fb' => 'IXFR',
	'00fc' => 'AXFR',
	'00ff' => 'ANY',
);

# global variables
my $VERSION = '0.2';
my $this_minute;
my %sum = (
	(map {$_ => 0} values %querytypes),
);
my $rrd_inited=0;
my %opt;

sub usage
{
	print "Usage: tinygraph [options]\n\n";
	print "  -h, --help       Display this help screen and exit\n";
	print "  -V, --version    Output version information and exit\n";
	print "  -c, --cat        Causes the logfile to be only read and not monitored\n";
	print "  -l, --logfile f  Use logfile f instead of $logfile\n";
	print "  -r, --rrd f      Use RRD file f instead of $rrd\n";

	exit(0);
}

sub version
{
	print "tinygraph $VERSION by Ulrich Zehl <ulrich\@topfen.net>\n\n";
	print "This program is provided under the terms of the GPL\n";
	print "and comes with absolutely NO WARRANTY.\n";

	exit(0);
}

sub process_line($)
{
	my $line = shift;
	if ($line =~ /^(@[a-f0-9]{24}) ([a-f0-9]{8}):[a-f0-9]{4}:[a-f0-9]{4} ([\+\-IC\/]) ([a-f0-9]{4}) (.+)$/)
	{
		my $time    = tai2unix($1);
		my $ip      = join(".", unpack("C*", pack("H8", $2)));
		my $rtype   = $3;
		my $qtype   = $querytypes{$4};
		#my $qstring = $5; # currently unused
	
		if ($rtype eq '+')
		{
			event($time, $qtype);
		}
	}
}

sub init_rrd($)
{
	my $m = shift;
	my $rows = $xpoints/$points_per_sample;
	my $realrows = int($rows*1.1); # ensure that the full range is covered
	my $day_steps = int(3600*24 / ($rrdstep*$rows));
	# use multiples, otherwise rrdtool could choose the wrong RRA
	my $week_steps = $day_steps*7;
	my $month_steps = $week_steps*5;
	my $year_steps = $month_steps*12;

	if (! -f $rrd)
	{
		RRDs::create($rrd, '--start', $m, '--step', $rrdstep,
				(map { "DS:$_:ABSOLUTE:".($rrdstep*2).':0:U' } values %querytypes),
				"RRA:AVERAGE:0.5:$day_steps:$realrows",   # day
				"RRA:AVERAGE:0.5:$week_steps:$realrows",  # week
				"RRA:AVERAGE:0.5:$month_steps:$realrows", # month
				"RRA:AVERAGE:0.5:$year_steps:$realrows",  # year
				"RRA:MAX:0.5:$day_steps:$realrows",   # day
				"RRA:MAX:0.5:$week_steps:$realrows",  # week
				"RRA:MAX:0.5:$month_steps:$realrows", # month
				"RRA:MAX:0.5:$year_steps:$realrows",  # year
		);
		$this_minute = $m;
	}
	else
	{
		$this_minute = RRDs::last($rrd) + $rrdstep;
	}

	$rrd_inited=1;
}

sub event($$)
{
	my ($t, $type) = @_;
	update($t) and $sum{$type}++;
}

# returns 1 if $sum should be updated
sub update($)
{
	my $t = shift;
	my $m = $t - $t%$rrdstep;
	init_rrd($m) unless $rrd_inited;
	return 1 if $m == $this_minute;
	return 0 if $m < $this_minute;

	my $update_line = "$m";
	foreach my $type (values %querytypes)
	{
		$update_line .= ":$sum{$type}";
		$sum{$type} = 0;
	}

	RRDs::update($rrd, $update_line);
	if($m > $this_minute+$rrdstep)
	{
		for(my $sm=$this_minute+$rrdstep; $sm<$m; $sm+=$rrdstep)
		{
			$update_line  = $sm;
			$update_line .= ':0' foreach (keys %querytypes);
			RRDs::update $rrd, $update_line;
		}
	}
	$this_minute = $m;
	return 1;
}

sub main()
{
	Getopt::Long::Configure('no_ignore_case');
	GetOptions(\%opt, 
		'help|h',
		'cat|c',
		'logfile|l=s',
		'version|V',
		'rrd|r=s'
	) or exit(1);

	usage()   if $opt{help};
	version() if $opt{version};

	$logfile = $opt{logfile} if defined $opt{logfile};
	$rrd = $opt{rrd} if defined $opt{rrd};

	if ($opt{cat})
	{
		open(LOG, "<$logfile")
			or die "Error opening $logfile: $!";

		while (<LOG>)
		{
			process_line($_);
		}
	}
	else
	{
		my $file = File::Tail->new(name => $logfile, tail => -1);

		while (my $line = $file->read)
		{
			process_line($line);
		}
	}

	exit(0);
}

main;
